from typing import Any
import os
import datetime
import cv2 as cv
import numpy as np
from scipy.stats import norm
import matplotlib.pyplot as plt
# -------------------------------------------------------------------------------------------
# from numba import vectorize, cuda, jit
# from multiprocessing import Pool
# import statistics as stat
# from PIL import Image, ImageChops
# from scipy.fftpack import fft, fftshift, ifft
# from numba import vectorize

def getusershift(filenamesample, numoffiles, skip, displayscale):

    pindex = filenamesample.index('.')
    inputfileprefix = filenamesample[0:pindex-4]
    ext = filenamesample[pindex:]

    img = cv.imread(filenamesample, cv.IMREAD_ANYDEPTH)
    numofrow = np.size(img, 0)
    numofcol = np.size(img, 1)

    mouseinputlist = []

    i = 1
    while i <= numoffiles:

        if  i < 10:
            inputfilename = inputfileprefix + '000' + str(i) + ext
        elif i < 100:
            inputfilename = inputfileprefix + '00' + str(i) + ext
        elif i < 1000:
            inputfilename = inputfileprefix + '0' + str(i) + ext
        else:
            inputfilename = inputfileprefix + str(i) + ext


        def mouseclick(event, mousex, mousey, flags, param):

            if event == cv.EVENT_LBUTTONDOWN:

                mouseinputlist.append((i, round(mousex/displayscale), round(mousey/displayscale)))

                cv.destroyAllWindows()

            return
 
        imgflat = cv.imread(inputfilename, cv.IMREAD_ANYDEPTH)
        # read the image in grayscale mode (imread & imwrite tranposes the array)


        imgscaled = cv.resize(imgflat, (round(numofcol*displayscale), round(numofrow*displayscale)))
        cv.imshow(inputfilename, imgscaled)
        # displayed the raw image
    
        cv.setMouseCallback(inputfilename, mouseclick)

        cv.waitKey(0)

        i = i + skip

    mouseinputarray = np.array(mouseinputlist)

    np.savetxt('mouseinput.csv', mouseinputarray, delimiter = ',', fmt = '%i')

    rowshift = int(max(mouseinputarray[:, 2])- min(mouseinputarray[:, 2]))
    colshift = int(max(mouseinputarray[:, 1])- min(mouseinputarray[:, 1]))
    print('row shift: ' + str(rowshift))
    print('column shift: ' + str(colshift))

    rowshiftmid = round((max(mouseinputarray[:, 2])+min(mouseinputarray[:, 2]))/2)
    colshiftmid = round((max(mouseinputarray[:, 1])+min(mouseinputarray[:, 1]))/2)
    print('row shift mid-point: ' + str(rowshiftmid))
    print('column shift mid-point: ' + str(colshiftmid))


    if skip > 1: # interpolation carried out

        [r6, r5, r4, r3, r2, r1, r0] = np.polyfit(mouseinputarray[:, 0], mouseinputarray[:, 2], 6)
        [c6, c5, c4, c3, c2, c1, c0] = np.polyfit(mouseinputarray[:, 0], mouseinputarray[:, 1], 6)
        
        absshiftinterpol = np.zeros((0, 2))

        n = 1
        while n <= numoffiles:

            colshiftcalc = int(round(c0+c1*(n)+c2*(n**2)+c3*(n**3)+c4*(n**4)+c5*(n**5)+c6*(n**6)))
            rowshiftcalc = int(round(r0+r1*(n)+r2*(n**2)+r3*(n**3)+r4*(n**4)+r5*(n**5)+r6*(n**6)))
            absshiftinterpol = np.vstack([absshiftinterpol, [colshiftcalc, rowshiftcalc]])

            n = n + 1
        
        rowshift = int(max(absshiftinterpol[:, 1])- min(absshiftinterpol[:, 1]))
        colshift = int(max(absshiftinterpol[:, 0])- min(absshiftinterpol[:, 0]))
        print('calculated row shift: ' + str(rowshift))
        print('calculated column shift: ' + str(colshift))

        rowshiftmid = round((max(absshiftinterpol[:, 1])+min(absshiftinterpol[:, 1]))/2)
        colshiftmid = round((max(absshiftinterpol[:, 0])+min(absshiftinterpol[:, 0]))/2)
        print('calculated row shift mid-point: ' + str(rowshiftmid))
        print('calculated column shift mid-point: ' + str(colshiftmid))
        # calculated the total shift flat

        usershift = np.zeros((0,2))
        n = 1
        while n <= numoffiles:

            usershift = np.vstack([usershift, [colshiftmid-absshiftinterpol[n-1, 0]-int(colshift/2) , rowshiftmid-absshiftinterpol[n-1, 1]-int(rowshift/2)]])

            n = n + 1

        usershift = usershift.astype('int')

        np.savetxt('usershift.csv', usershift, delimiter = ',', fmt = '%i')


    elif skip == 1: # interpolation skipped

        usershift = np.zeros((0,2))
        n = 1
        while n <= numoffiles:

            usershift = np.vstack([usershift, [colshiftmid-absshiftarray[n-1, 2]-int(colshiftmid/2) , rowshiftmid-absshiftarray[n-1, 1]-int(rowshiftmid/2)]])

            n = n + 1

        usershift = usershift.astype('int')

        np.savetxt('usershift.csv', usershift, delimiter = ',', fmt = '%i')

    print("Task completed!")

    return

def volread(filenamesample, numoffiles):

    pindex = filenamesample.index('.')
    inputfileprefix = filenamesample[0:pindex-4]
    ext = filenamesample[pindex:]

    print('reading the image sequence...')

    numofrow = np.size(cv.imread(filenamesample), 0)
    numofcol = np.size(cv.imread(filenamesample), 1)
    numofslice = numoffiles

    volume = np.zeros((numofrow, numofcol, numofslice))

    i = 1
    while i <= numoffiles:

        if  i < 10:
            inputfilename = inputfileprefix + '000' + str(i) + ext
        elif i < 100:
            inputfilename = inputfileprefix + '00' + str(i) + ext
        elif i < 1000:
            inputfilename = inputfileprefix + '0' + str(i) + ext
        else:
            inputfilename = inputfileprefix + str(i) + ext

        volume[:, :, i-1] = cv.imread(inputfilename, cv.IMREAD_ANYDEPTH)

        i = i + 1

    print('image sequence successfully read into a 3d array')

    return volume

def volstat(volume):

    numofrow = np.size(volume, 0)
    numofcol = np.size(volume, 1)
    numofslice = np.size(volume, 2)

    fig, volstatplot = plt.subplots(3,2)
    fig.suptitle('volume statistics')

    rowstat = []
    
    i = 1
    while i <= numofrow:

        imgflat = volume[i-1, :, :]
        (mode, sdev) = norm.fit(imgflat)
        rowstat.append((int(i), int(mode), int(sdev)))
        i = i + 1

    rowstatarray = np.array(rowstat)
    np.savetxt('rowstat.csv', rowstatarray, delimiter=',', fmt='%i')

    volstatplot[0,0].plot(rowstatarray[:,0], rowstatarray[:,1])
    volstatplot[0,0].set_ylabel('row')
    volstatplot[0,1].plot(rowstatarray[:,0], rowstatarray[:,2])

    colstat = []

    i = 1
    while i <= numofcol:

        imgflat = volume[:, i-1, :]
        (mode, sdev) = norm.fit(imgflat)
        colstat.append((int(i), int(mode), int(sdev)))
        i = i + 1

    colstatarray = np.array(colstat)
    np.savetxt('colstat.csv', colstatarray, delimiter=',', fmt='%i')

    volstatplot[1,0].plot(colstatarray[:,0], colstatarray[:,1])
    volstatplot[1,0].set_ylabel('column')
    volstatplot[1,1].plot(colstatarray[:,0], colstatarray[:,2])

    slicestat = []

    i = 1
    while i <= numofslice:

        imgflat = volume[:, :, i-1]
        (mode, sdev) = norm.fit(imgflat)
        slicestat.append((int(i), int(mode), int(sdev)))
        i = i + 1

    slicestatarray = np.array(slicestat)
    np.savetxt('slicestat.csv', slicestatarray, delimiter=',', fmt='%i')

    volstatplot[2,0].plot(slicestatarray[:,0], slicestatarray[:,1])
    volstatplot[2,0].set_xlabel('mode')
    volstatplot[2,0].set_ylabel('slice')
    volstatplot[2,1].plot(slicestatarray[:,0], slicestatarray[:,2])
    volstatplot[2,1].set_xlabel('standard deviation')

    print('Task complete! volume statistics exported as csv files to the current folder')
    plt.show(fig)

    return

def volsmooth(volume, px):

    numofrow = np.size(volume, 0)
    numofcol = np.size(volume, 1)
    numofslice = np.size(volume, 2)

    output = np.zeros((numofrow, numofcol, numofslice))

    i = 0 #rownumber
    j = 0 #colnumber
    k = 0 #slicenumber

    while k < numofslice:

        if k <= px:
            zs = 0
            ze = k + px
        elif k > px and k <= numofslice-px:
            zs = k - px
            ze = k + px
        else:
            zs = k - px
            ze = numofslice

        while i < numofrow:
            if i <= px:
                xs = 0
                xe = i + px
            elif i > px and i <= numofrow-px:
                xs = i - px
                xe = i + px
            else:
                xs = i - px
                xe = numofrow
            while j < numofcol:
                if j <= px:
                    ys = 0
                    ye = j + px
                elif j > px and j <= numofcol-px:
                    ys = j - px
                    ye = j + px
                else:
                    ys = j - px
                    ye = numofcol
                # subarray indices calculated

                subarray = np.array(volume[xs:xe, ys:ye, zs:ze], dtype=np.int64)
                output[i, j, k] = int((np.sum(np.sum(np.sum(subarray, axis=2), axis=1)))/((ze-zs)*(ye - ys)*(xe - xs)))

                j += 1
            i += 1
            j = 0

        output = output.astype('uint16')

        sourcedir = str(os.getcwd())
        os.makedirs('volsmooth-'+str(px)+'px', exist_ok=True)
        os.chdir(sourcedir + '/volsmooth-'+str(px)+'px/')

        if  k+1 < 10:
            ofn = 'vs-' + '000' + str(k+1) + '.tif'
        elif k+1 < 100:
            ofn = 'vs-' + '00' + str(k+1) + '.tif'
        elif k+1 < 1000:
            ofn = 'vs-' + '0' + str(k+1) + '.tif'
        else:
            ofn = 'vs-' + str(k+1) + '.tif'

        cv.imwrite(ofn, output[:, :, k])

        os.chdir(sourcedir)

        print(f'slice {k+1} exported @ {datetime.datetime.now()}')

        k += 1
        i = 0
        j = 0

    return

def volnorm(volume, llf, ulf):

    numofrow = np.size(volume, 0)
    numofcol = np.size(volume, 1)
    numofslice = np.size(volume, 2)

    print('normalization initiated...')
    
    i = 1
    while i <= numofslice:

        (mode, sdev) = norm.fit(volume[ :, :, i-1])
        ll = mode - sdev*llf
        volume[:, :, i-1] = (volume[:, :, i-1] - ll)*(65535/((llf+ulf)*sdev))

        i = i + 1

    i = 1
    while i <= numofrow:

        (mode, sdev) = norm.fit(volume[i-1, :, :])
        ll = mode - sdev*llf
        volume[i-1, :, :] = (volume[i-1, :, :] - ll)*(65535/((llf+ulf)*sdev))

        i = i + 1

    i = 1
    while i <= numofcol:

        (mode, sdev) = norm.fit(volume[ :, i-1, :])
        ll = mode - sdev*llf
        volume[:, i-1, :] = (volume[:, i-1, :] - ll)*(65535/((llf+ulf)*sdev))

        i = i + 1

    print('volume normalized along all three axes! removal of outliers initiated...')

    l = 0

    i = 0
    while i < numofrow:
        j = 0
        while j < numofcol:
            k = 0
            while k < numofslice:

                if volume[i, j, k] < 0:
                    volume[i, j, k] = int(0)
                elif volume[i,j,k] > 65535:
                    volume[i,j,k] = int(65535)
                else:
                    volume[i,j,k] = int(volume[i,j,k])

                k += 1
            j = j +1
            k = 0
        i = i + 1
        j = 0
        k = 0

        if l != int(i/(numofrow*0.01)):
            l = int(i/(numofrow*0.01))
            print(f'{l} percent complete.. @ {datetime.datetime.now()}')

    volume = volume.astype('uint16')

    print('outliers removed! file export initiated...')

    sourcedir = str(os.getcwd())
    os.makedirs('volnorm', exist_ok=True)
    os.chdir(sourcedir + '/volnorm/')
    # created the output folder

    i = 1
    while  i <= numofslice:
        if  i < 10:
            filename = '000' + str(i) + '.tif'
        elif i < 100:
            filename = '00' + str(i) + '.tif'
        elif i < 1000:
            filename = '0' + str(i) + '.tif'
        else:
            filename = str(i) + '.tif'

        cv.imwrite('volnorm-' + filename, volume[:,:,i-1])

        i += 1
    # wrote the output images in the output folder

    os.chdir(sourcedir)
    # changed working directory back to the source

    print('Task complete! files exported to /volnorm/ subfolder')

    return

def porosity(volume, thresh):

    numofrow = np.size(volume, 0)
    numofcol = np.size(volume, 1)
    numofslice = np.size(volume, 2)

    print('porosity calculation initiated...')
    print('slice, porosity')
    porositylist = []

    i = 0
    while i < numofslice:

        thisslice = volume[:, :, i]

        numofrow = np.size(thisslice, 0)
        numofcol = np.size(thisslice, 1)

        thisslice[thisslice < thresh] = 0
        thisslice[thisslice >= thresh] = 255
        

        porosity = int((np.sum(np.sum(thisslice, axis=1)))/(numofrow*numofrow*2.55))

        porositylist.append(porosity)

        print(f'{i+1}, {porosity}%')
        
        i = i + 1

    porosityarray = np.array(porositylist)

    avgporosity = int(np.sum(porosityarray)/numofslice)

    print(f'porosity of the volume: {avgporosity}%')

    return 

def colormap(filenamesample, numoffiles, colormap):

    pindex = filenamesample.index('.')
    inputfileprefix = filenamesample[0:pindex-4]
    ext = filenamesample[pindex:]

    i = 1
    while i <= numoffiles:

        if  i < 10:
            inputfilename = inputfileprefix + '000' + str(i) + ext
        elif i < 100:
            inputfilename = inputfileprefix + '00' + str(i) + ext
        elif i < 1000:
            inputfilename = inputfileprefix + '0' + str(i) + ext
        else:
            inputfilename = inputfileprefix + str(i) + ext

        image = cv.imread(inputfilename, cv.IMREAD_ANYDEPTH)
        # img8bit = image.astype('uint8')

        plt.imsave('cmap'+inputfilename[0:-4]+'.tiff' , image, vmin=0, vmax=65535, cmap = colormap)

        i = i + 1

    print('Task complete! files exported to the current folder')

    return


# to do list: upgrade all cv.imread to anydepth; upgrade all numpy sum arrays to int64; 
# fix the multiprocess
# ---------------------------------------------------------------------------------------------
# ---------------------------------------------------------------------------------------------
#  ------------------------------UNDER DEVELOPMENT --------------------------------------------
# ---------------------------------------------------------------------------------------------
# ---------------------------------------------------------------------------------------------


# ============================================================================================
#                FUNCTION TO FIND AXIS OF ROTATION OF ALIGNED XCT PROJECTIONS                #
# ============================================================================================

def findaorinputlist(inputsinogram, totaltheta, aorll, aorul, pixelinterlace, findaorprjs):

    inputlist = []

    i = aorll
    while i <= aorul:

        inputlist.append((inputsinogram, totaltheta, i, findaorprjs))

        i = i + pixelinterlace

    print('multiprocessing inputlist generated')
    print('finding axis of rotation initiated...')


    return inputlist

def findaor(inputsinogram, totaltheta, aorshift, findaorprjs):

    i = aorshift

    imgflat = ((np.sum(cv.imread(inputsinogram), axis=2)/3))
    numofprojection = np.size(imgflat, 1)
    numofrow = np.size(imgflat, 0)
    scale = findaorprjs/numofprojection
    imgscaled = cv.resize(imgflat, (round(numofprojection*scale), numofrow))
    numofprojection = np.size(imgscaled, 1)

    output = np.zeros((numofrow, numofrow))

    thetalist = []
    j = 1

    while j <= numofprojection:
        thetalist.append((totaltheta*(j-1))/(numofprojection-1))
        j = j + 1

    theta = np.array(thetalist)

    width = min(numofrow-i, i)
    xs = i-width
    xe = i+width            
    sinogram = imgscaled[xs:xe, :]
    # ------------------------------------------------------------------------------------------------------------
    #                                           FUNCTION TO BACK-PROJECTION                                      #
    # inputs:  sinogram - [n x m] numpy array where n is the number of projections and m the number of angles    #
    # theta - vector of length m denoting the angles represented in the sinogram                                 #
    # output: backprojArray - [n x n] backprojected 2-D numpy array"""                                           #
    #                               credit: https://github.com/kyochujoho/filt-back-proj                         #
    # ------------------------------------------------------------------------------------------------------------
    imageLen = sinogram.shape[0]
    reconMatrix = np.zeros((imageLen, imageLen))
    
    x = np.arange(imageLen)-imageLen/2 #create coordinate system centered at (x,y = 0,0)
    y = x.copy()
    X, Y = np.meshgrid(x, y)

    plt.ion()
    fig2, ax = plt.subplots()
    im = plt.imshow(reconMatrix, cmap='gray')

    theta = theta*np.pi/180
    numAngles = len(theta)

    for n in range(numAngles):
        Xrot = X*np.sin(theta[n])-Y*np.cos(theta[n]) #determine rotated x-coordinate about origin in mesh grid form
        XrotCor = np.round(Xrot+imageLen/2) #shift back to original image coordinates, round values to make indices
        XrotCor = XrotCor.astype('int')
        projMatrix = np.zeros((imageLen, imageLen))
        m0, m1 = np.where((XrotCor >= 0) & (XrotCor <= (imageLen-1))) #after rotating, you'll inevitably have new coordinates that exceed the size of the original
        s = sinogram[:,n] #get projection
        projMatrix[m0, m1] = s[XrotCor[m0, m1]]  #backproject in-bounds data
        reconMatrix += projMatrix
        im.set_data(Image.fromarray((reconMatrix-np.min(reconMatrix))/np.ptp(reconMatrix)*255))
        ax.set_title('Theta = %.2f degrees' % (theta[n]*180/np.pi))
        fig2.canvas.draw()
        fig2.canvas.flush_events()
     
    plt.close()
    plt.ioff()
    
    backprojArray = np.flipud(reconMatrix)

    xs = round((numofrow-2*width)/2)
    ys = xs
    xe = np.size(backprojArray, 0)+xs
    ye = xe
    output[xs:xe, ys:ye] = backprojArray

    output = output.astype('uint16')

    print(np.size(backprojArray))
    print(np.size(output[xs:xe, ys:ye]))

    sourcedir = str(os.getcwd())
    os.makedirs('findaor', exist_ok=True)
    os.chdir(sourcedir + '/findaor/')
    cv.imwrite(str(i)+'px-' + inputsinogram[0:6] + '.tif', output) 
    os.chdir(sourcedir)

    print(str(i)+'px-' + inputsinogram[0:6] + '.tif' + ', exported to /findaor/ at ' + str((datetime.datetime.now())))

    return

# =============================================================================================
#                             GET IMAGE STATISTICS OF GRAYSCALE IMAGES                        #
# =============================================================================================

def getimagestat(filenamesample, numoffiles):
    
    pindex = filenamesample.index('.')
    inputfileprefix = filenamesample[0:pindex-4]
    ext = filenamesample[pindex:]

    img = cv.imread(filenamesample, cv.IMREAD_ANYDEPTH)
    numofrow = np.size(img, 0)
    numofcol = np.size(img, 1)

    imagestat  = np.zeros((0,5))

    print('file #; mode; std dev; min; max')

    i = 1
    while i <= numoffiles:

        if  i < 10:
            inputfilename = inputfileprefix + '000' + str(i) + ext
        elif i < 100:
            inputfilename = inputfileprefix + '00' + str(i) + ext
        elif i < 1000:
            inputfilename = inputfileprefix + '0' + str(i) + ext
        else:
            inputfilename = inputfileprefix + str(i) + ext

        img  = cv.imread(inputfilename, cv.IMREAD_ANYDEPTH)

        (mode, sdev) = norm.fit(img)
        (minimum, maximum) = (np.amin(img), np.amax(img))
        imagestat = np.vstack([imagestat, (i, mode, sdev, minimum, maximum)])

        print(imagestat[i-1, :])

        i = i + 1
    
    np.savetxt('imagestat.csv', imagestat, delimiter = ',', header='file #; mode; std dev; min; max')

    print('Task complete, imagestat.csv exported!')

    return imagestat

# ============================================================================================
#           NORMALIZING REGIONAL IRREGULARITIES IN INTENSITY OF GRAYSCALE IMAGES             #
# ============================================================================================
def regnorminputlist(filenamesample, numoffiles, px):
    
    pindex = filenamesample.index('.')
    inputfileprefix = filenamesample[0:pindex-4]
    ext = filenamesample[pindex:]

    inputlist = []

    i = 1
    while i <= numoffiles:

        if  i < 10:
            inputfilename = inputfileprefix + '000' + str(i) + ext
        elif i < 100:
            inputfilename = inputfileprefix + '00' + str(i) + ext
        elif i < 1000:
            inputfilename = inputfileprefix + '0' + str(i) + ext
        else:
            inputfilename = inputfileprefix + str(i) + ext

        inputlist.append([inputfilename, px])

        i = i + 1

    print('multiprocessing inputlist generated')

    return inputlist

def regnorm(inputfilename, px):

    imgflat = np.array(cv.imread(inputfilename, cv.IMREAD_ANYDEPTH), dtype=np.int64)
    # read the image in grayscale mode (imread & imwrite tranposes the array)

    numofrow = np.size(imgflat, 0)
    numofcol = np.size(imgflat, 1)
    # extracted the dimensions of the image

    correctionflat = np.array(np.zeros((numofrow, numofcol)), dtype=np.int64)
    outputflat = np.array(np.zeros((numofrow, numofcol)), dtype=np.int64)
    # initated the indicies for the correction array

    i = 0 #rownumber
    j = 0 #colnumber

    while i < numofrow:
        if i <= px:
            xs = 0
            xe = i + px
        elif i > px and i <= numofrow-px:
            xs = i - px
            xe = i + px
        else:
            xs = i - px
            xe = numofrow
        while j < numofcol:
            if j <= px:
                ys = 0
                ye = j + px
            elif j > px and j <= numofcol-px:
                ys = j - px
                ye = j + px
            else:
                ys = j - px
                ye = numofcol
            # subarray indices calculated
            
            subarray = np.array(imgflat[xs:xe, ys:ye], dtype=np.int64)
            correctionflat[i, j] = (np.sum(np.sum(subarray, axis=1)))/((ye - ys)*(xe - xs))
            
            j += 1
        i += 1
        j = 0

    (mu, sigma) = norm.fit(correctionflat)

    outputflat = imgflat - correctionflat + mu

    outputflat = outputflat.astype('uint16')
    correctionflat = correctionflat.astype('uint16')
    # carried out the correction

    sourcedir = str(os.getcwd())
    os.makedirs('regnorm', exist_ok=True)
    os.chdir(sourcedir + '/regnorm/')
    # created the output folder

    cfn = 'correction-' + str(px) + 'px-' + inputfilename
    ofn = 'output-' + str(px) + 'px-' + inputfilename
    # created output files

    cv.imwrite(cfn, correctionflat)
    cv.imwrite(ofn, outputflat)
    # wrote the output images in the output folder

    os.chdir(sourcedir)
    # changed working directory back to the source

    print('{ofn} processed')

    return

# ============================================================================================
#                        NORMALIZING A BATCH OF GRAY 16 BIT IMAGES                           #
# ============================================================================================

def batchnorminputlist(filenamesample, numoffiles, factor):

    pindex = filenamesample.index('.')
    inputfileprefix = filenamesample[0:pindex-4]
    ext = filenamesample[pindex:]

    inputlist = []

    i = 1
    while i <= numoffiles:

        if  i < 10:
            inputfilename = inputfileprefix + '000' + str(i) + ext
        elif i < 100:
            inputfilename = inputfileprefix + '00' + str(i) + ext
        elif i < 1000:
            inputfilename = inputfileprefix + '0' + str(i) + ext
        else:
            inputfilename = inputfileprefix + str(i) + ext

        inputlist.append((inputfilename, factor))

        i = i + 1
    
    print('multiprocessing inputlist generated')

    return inputlist

def batchnorm(inputfilename, factor):

    imgflat  = cv.imread(inputfilename, cv.IMREAD_ANYDEPTH)
    numofrow = np.size(imgflat, 0)
    numofcol = np.size(imgflat, 1)
    
    (mode, sdev) = norm.fit(imgflat)

    ll = mode - sdev*factor
    
    outputtemp = (imgflat - ll)*(65535/(2*factor*sdev))

    outputflat = np.zeros((numofrow, numofcol))

    i = 0
    while i < numofrow:
        j = 0
        while j < numofcol:

            if outputtemp[i, j] < 0:
                outputflat[i, j] = int(0)
            elif outputtemp[i,j] > 65535:
                outputflat[i,j] = int(65535)
            else:
                outputflat[i, j] = int(outputtemp[i,j])

            j = j +1
        i = i + 1
        j=0

    outputflat = outputflat.astype('uint16')

    sourcedir = str(os.getcwd())
    os.makedirs('batchnorm', exist_ok=True)
    os.chdir(sourcedir + '/batchnorm/')

    cv.imwrite(inputfilename, outputflat)
    os.chdir(sourcedir)

    return

# ============================================================================================
#                              FUNCTIONS TO ALIGN XCT PROJECTIONS                            #
# ============================================================================================

def alignprojections(technique, filenamesample, numoffiles, skip, param1):

    # ('manual', 'ncs55nm0001.tif', 801, 40, 0.40)

    pindex = filenamesample.index('.')
    inputfileprefix = filenamesample[0:pindex-4]
    ext = filenamesample[pindex:]
    outputfileprefix = inputfileprefix[0:6]+'-'

    # ---------------------------------------------------------------------------------------------
    #                       CREATES AN ARRAY OF INPUT/OUTPUT FILENAMES                            #
    # ---------------------------------------------------------------------------------------------

    filenamelist = []

    i = start
    while i <= end:

        if  i < 10:
            inputfilename = inputfileprefix + '000' + str(i) + ext
            outputfilename = outputfileprefix + '000' + str(i) + ext
        elif i < 100:
            inputfilename = inputfileprefix + '00' + str(i) + ext
            outputfilename = outputfileprefix + '00' + str(i) + ext
        elif i < 1000:
            inputfilename = inputfileprefix + '0' + str(i) + ext
            outputfilename = outputfileprefix + '0' + str(i) + ext
        else:
            inputfilename = inputfileprefix + str(i) + ext
            outputfilename = outputfileprefix + str(i) + ext

        filenamelist.append([inputfilename, outputfilename])

        i = i + 1
    
    filenamearray = np.array(filenamelist)

    img = cv.imread(filenamearray[0, 0])
    numofrow = np.size(img, 0)
    numofcol = np.size(img, 1)
    numofch = np.size(img, 2)

    # ---------------------------------------------------------------------------------------------
    #                   MANUAL TRANSLATIONAL ALIGNMENT OF XCT PROJECTIONS                         #
    # ---------------------------------------------------------------------------------------------

    if technique == 'manual':

        print('Manual alignment initiated...')

        displayscale = param1
        absshiftlist = []

        print('please left click on the reference feature')

        i = start

        while i <= end:

            def mouseclick(event, mousex, mousey, flags, param):

                if event == cv.EVENT_LBUTTONDOWN:

                    absshiftlist.append([i, round(mousey/displayscale), round(mousex/displayscale), 0])
                    cv.destroyAllWindows()

                return

            img = cv.imread(filenamearray[i-start, 0])
            imgflat = np.sum(img, axis=2)/3
            # read the image in grayscale mode (imread & imwrite tranposes the array)

            imgscaled = cv.resize(img, (round(numofcol*displayscale), round(numofrow*displayscale)))
            cv.imshow(filenamearray[i-start, 0], imgscaled)
            # displayed the raw image
        
            cv.setMouseCallback(filenamearray[i-start, 0], mouseclick)

            cv.waitKey(0)

            i = i + fileinterlace

        absshiftarray = np.array(absshiftlist)
        sourcedir = str(os.getcwd())
        os.makedirs('AlignProjections', exist_ok=True)
        os.chdir(sourcedir + '/AlignProjections/')
        np.savetxt(outputfileprefix+'-manual-absshiftarray.csv', absshiftarray, delimiter = ',')
        os.chdir(sourcedir)

    # ---------------------------------------------------------------------------------------------
    #                     CORRELATION BASED ALIGNMENT OF XCT PROJECTIONS                          #
    # ---------------------------------------------------------------------------------------------

    elif technique == 'correlation':

        print('Correlation based alignment initiated...')

        pixelshiftperframe = param1
        pixelinterlace = param2
        px = pixelshiftperframe
        intr = pixelinterlace
        relshiftrecord = np.zeros((1,4))
        absshiftarray = np.zeros((1,4))

        i = start + fileinterlace

        while i <= end:

            img = cv.imread(filenamearray[i-start, 0])
            imgflat = np.sum(img, axis=2)/3
            ref = cv.imread(filenamearray[i-start-fileinterlace, 0])
            refflat = np.sum(ref, axis=2)/3
            print('Processing, ' + filenamearray[i-start, 0] + ', started at ' + str(datetime.datetime.now()))
            print('reference filename: ' + filenamearray[i-start-fileinterlace, 0])

            k = 0 #rownum
            l = 0 #colnum
            while k <= px:
                while l <= px:
                    if k == 0 and l == 0:
                        row = 0
                        col = 0
                        croppedimg = imgflat
                        croppedref = refflat
                        cc = np.sum(np.sum(np.absolute(croppedimg-croppedref), axis=1))/((numofrow)*(numofcol))
                        # ccrecord = np.vstack([ccrecord, (row, col, cc)])
                        ccrecord = [i, row, col, cc]
                    else:
                        row = -k
                        col = -l
                        croppedimg = imgflat[k:numofrow+1, l:numofcol+1]
                        croppedref = refflat[0:numofrow-k, 0:numofcol-l]
                        cc = np.sum(np.sum(np.absolute(croppedimg-croppedref), axis=1))/((numofrow-k)*(numofcol-l))
                        ccrecord = np.vstack([ccrecord, (i, row, col, cc)])
                        row = k
                        col = l
                        croppedimg = imgflat[0:numofrow-k, 0:numofcol-l]
                        croppedref = refflat[k:numofrow+1, l:numofcol+1]
                        cc = np.sum(np.sum(np.absolute(croppedimg-croppedref), axis=1))/((numofrow-k)*(numofcol-l))
                        ccrecord = np.vstack([ccrecord, (i, row, col, cc)])
                    l = l + intr
                k = k + intr
                l = 0              
            # generated correlation coefficient tables

            relshift = ccrecord[np.where(ccrecord[:,3] == min(ccrecord[:,3]))[0][0], 0:4]
            relshiftrecord = np.vstack([relshiftrecord, relshift])
            absshiftarray = np.vstack([absshiftarray, [i, np.sum(relshiftrecord[:, 1]), np.sum(relshiftrecord[:, 2]), relshiftrecord[-1, 3]]])
            # calcualted absolute shift from relative shift

            sourcedir = str(os.getcwd())
            os.makedirs('AlignProjections', exist_ok=True)
            os.chdir(sourcedir + '/AlignProjections/')
            np.savetxt(outputfileprefix+'corr-relshiftarray.csv', relshiftrecord, delimiter = ',')
            np.savetxt(outputfileprefix+'corr-absshiftarray.csv', absshiftarray, delimiter = ',')
            os.chdir(sourcedir)


            i = i + fileinterlace
            #moved to the next image

    # ---------------------------------------------------------------------------------------------
    #                  COUNT-DISTRIBUTION BASED ALIGNMENT OF XCT PROJECTIONS                      #
    # ---------------------------------------------------------------------------------------------

    elif technique == 'countdist':
        print('Count distribution based alignment initiated...')


    else:
        print('Invalid alignment technique choice')

    # ---------------------------------------------------------------------------------------------
    #                        COMMUNICATES MEASURED SHIFT INFORMATION                              #
    # ---------------------------------------------------------------------------------------------

    rowshift = int(max(absshiftarray[:, 1])- min(absshiftarray[:, 1]))
    colshift = int(max(absshiftarray[:, 2])- min(absshiftarray[:, 2]))
    print('row shift: ' + str(rowshift))
    print('column shift: ' + str(colshift))

    rowshiftmid = round((max(absshiftarray[:, 1])+min(absshiftarray[:, 1]))/2)
    colshiftmid = round((max(absshiftarray[:, 2])+min(absshiftarray[:, 2]))/2)
    print('row shift mid-point: ' + str(rowshiftmid))
    print('column shift mid-point: ' + str(colshiftmid))

    # ---------------------------------------------------------------------------------------------
    #                  OPTIONAL INTERPOLATION OF SHIFTS ACROSS ALL IMAGES                         #
    # ---------------------------------------------------------------------------------------------

    if fileinterlace > 1: # interpolation carried out

        [r6, r5, r4, r3, r2, r1, r0] = np.polyfit(absshiftarray[:, 0], absshiftarray[:, 1], 6)
        [c6, c5, c4, c3, c2, c1, c0] = np.polyfit(absshiftarray[:, 0], absshiftarray[:, 2], 6)
        
        absshiftinterpol = np.zeros((0, 2))

        n = start
        while n <= end :
            absshiftinterpol = np.vstack([absshiftinterpol, [int(round(r0+r1*(n)+r2*(n**2)+r3*(n**3)+r4*(n**4)+r5*(n**5)+r6*(n**6))), int(round(c0+c1*(n)+c2*(n**2)+c3*(n**3)+c4*(n**4)+c5*(n**5)+c6*(n**6)))]])
            n = n + 1
        
        rowshift = int(max(absshiftinterpol[:, 0])- min(absshiftinterpol[:, 0]))
        colshift = int(max(absshiftinterpol[:, 1])- min(absshiftinterpol[:, 1]))
        print('calculated row shift: ' + str(rowshift))
        print('calculated column shift: ' + str(colshift))

        rowshiftmid = round((max(absshiftinterpol[:, 0])+min(absshiftinterpol[:, 0]))/2)
        colshiftmid = round((max(absshiftinterpol[:, 1])+min(absshiftinterpol[:, 1]))/2)
        print('calculated row shift mid-point: ' + str(rowshiftmid))
        print('calculated column shift mid-point: ' + str(colshiftmid))
        # calculated the total shift flat

        sourcedir = str(os.getcwd())
        os.makedirs('AlignProjections', exist_ok=True)
        os.chdir(sourcedir + '/AlignProjections/')
        np.savetxt(outputfileprefix+'absshiftinterpol.csv', absshiftinterpol, delimiter = ',')
        os.chdir(sourcedir)

        absshiftfinal = absshiftinterpol

    elif fileinterlace == 1: # interpolation skipped

        absshiftfinal = absshiftarray[:, 1:3]

    # ---------------------------------------------------------------------------------------------
    #                               GENERATION OF ALIGNED IMAGES                                  #
    # ---------------------------------------------------------------------------------------------

    i = start

    while i <= end:

        img = cv.imread(filenamearray[i-start, 0])
        # read the image in grayscale mode (imread & imwrite tranposes the array)

        output = np.zeros((numofrow+rowshift+2, numofcol+colshift+2, numofch))
        # created ouput shell array

        if technique == 'correlation':
            rows = int((rowshift/2) + absshiftfinal[int((i-start)), 0] - rowshiftmid)
            cols = int((colshift/2) + absshiftfinal[int((i-start)), 1] - colshiftmid)
        else:
            rows = int((rowshift/2) - absshiftfinal[int((i-start)), 0] + rowshiftmid)
            cols = int((colshift/2) - absshiftfinal[int((i-start)), 1] + colshiftmid)

        output[rows:rows+numofrow, cols:cols+numofcol, 0:3] = img
        # generated the output image

        croppedoutput = output[rowshift:numofrow, colshift:numofcol, 0:3]
        croppedoutput = croppedoutput.astype('uint16')

        ofn = filenamearray[i-start, 1]

        sourcedir = str(os.getcwd())
        os.makedirs('AlignProjections', exist_ok=True)
        os.chdir(sourcedir + '/AlignProjections/')
        cv.imwrite(ofn, croppedoutput)
        os.chdir(sourcedir)

        print(ofn + " exported to /AlignProjections/. at " + str(datetime.datetime.now()))
        # printed progress

        i = i +1
        # moved to read the next image

    print("Task completed!")

    return

# =============================================================================================
#                  REMOVE NOISE THROUGH AVERAGING PIXEL COUNTS ACROSS FRAMES                  #
# =============================================================================================

def seqsmoothinputlist(filenamesample, outputfileprefix, numoffiles):
    
    pindex = filenamesample.index('.')
    inputfileprefix = filenamesample[0:pindex-4]
    ext = filenamesample[pindex:]
    # outputfileprefix = inputfileprefix[0:6]+'-'

    filenamelist = []

    i = 1
    while i <= numoffiles:

        if  i < 10:
            inputfilename = inputfileprefix + '000' + str(i) + ext
            outputfilename = outputfileprefix + '000' + str(i) + ext
        elif i < 100:
            inputfilename = inputfileprefix + '00' + str(i) + ext
            outputfilename = outputfileprefix + '00' + str(i) + ext
        elif i < 1000:
            inputfilename = inputfileprefix + '0' + str(i) + ext
            outputfilename = outputfileprefix + '0' + str(i) + ext
        else:
            inputfilename = inputfileprefix + str(i) + ext
            outputfilename = outputfileprefix + str(i) + ext

        filenamelist.append([inputfilename, outputfilename])

        i = i + 1
    
    filenamearray = np.array(filenamelist)

    inputlist = []

    fa = filenamearray

    i = 0
    while i <= numoffiles-1:

        if i+1 == numoffiles:
            inputlist.append(( fa[i-1, 0], fa[i, 0], fa[0, 0], fa[i, 1]))
        else:
            inputlist.append(( fa[i-1, 0], fa[i, 0], fa[i+1, 0], fa[i, 1]))

        i = i + 1

    print('multiprocessing inputlist generated')

    return inputlist

def seqsmooth(input1, input2, input3, output):

    img1flat = cv.imread(input1, cv.IMREAD_ANYDEPTH)
    img2flat = cv.imread(input2, cv.IMREAD_ANYDEPTH)
    img3flat = cv.imread(input3, cv.IMREAD_ANYDEPTH)


    img4flat = (0.5*img1flat + img2flat + 0.5*img3flat )/2

    img4flat = img4flat.astype('uint16')

    sourcedir = str(os.getcwd())
    os.makedirs('seqsmooth', exist_ok=True)
    os.chdir(sourcedir + '/seqsmooth/')
    cv.imwrite(output, img4flat)
    os.chdir(sourcedir)

    print(output + ' written into the subfolder /seqsmooth/ at: ' + str(datetime.datetime.now))

    return

